# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from awkward._v2.operations.reducers.ak_count import count  # noqa: F401
from awkward._v2.operations.reducers.ak_count_nonzero import count_nonzero  # noqa: F401
from awkward._v2.operations.reducers.ak_sum import sum  # noqa: F401
from awkward._v2.operations.reducers.ak_prod import prod  # noqa: F401
from awkward._v2.operations.reducers.ak_any import any  # noqa: F401
from awkward._v2.operations.reducers.ak_all import all  # noqa: F401
from awkward._v2.operations.reducers.ak_min import min  # noqa: F401
from awkward._v2.operations.reducers.ak_max import max  # noqa: F401
from awkward._v2.operations.reducers.ak_ptp import ptp  # noqa: F401
from awkward._v2.operations.reducers.ak_argmin import argmin  # noqa: F401
from awkward._v2.operations.reducers.ak_argmax import argmax  # noqa: F401
from awkward._v2.operations.reducers.ak_moment import moment  # noqa: F401
from awkward._v2.operations.reducers.ak_mean import mean  # noqa: F401
from awkward._v2.operations.reducers.ak_var import var  # noqa: F401
from awkward._v2.operations.reducers.ak_std import std  # noqa: F401
from awkward._v2.operations.reducers.ak_covar import covar  # noqa: F401
from awkward._v2.operations.reducers.ak_corr import corr  # noqa: F401
from awkward._v2.operations.reducers.ak_linear_fit import linear_fit  # noqa: F401
from awkward._v2.operations.reducers.ak_softmax import softmax  # noqa: F401
